/*
 * 
 * assembler.c - This file contains the routines necessary to convert mnemonics to opcodes as well as
 * interpret FUSE settings
 * 
 * Kris Bahnsen  <kbahnsen@gmail.com>
 * Copyright (C) 2007 Kris Bahnsen
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 */


#define TRUE 1
#define FALSE 0

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "sxhexout.h"
#include "assembler.h"


int read_asm(char *asm_file_in)
{
	int temp;
	int x;
	FILE *asm_file;
	FILE *tmpout;
//	FILE * hexfile;
    char linemap[1024][256];
	
	for(temp = 0;temp <= 1023;temp++)
	{
		for(x = 0; x <= 255;x++)
		{
			linemap[temp][x] = 0x00;
		}
	}
	asm_file = fopen(asm_file_in,"r");
	tmpout = fopen("in.tmp","w");
	
	if(asm_file == NULL || tmpout == NULL)
	{
		fprintf(stderr,"Error opening file\n");
		exit(1);
	}
	
	for(;tmpchar != EOF;)
	{
		tmpchar = fgetc(asm_file);
		
		if((int)tmpchar >= 0x61 && (int)tmpchar <= 0x7A)
		{ 
			
			tmpchar -= 0x20;
		}
		if(tmpchar != 0xFFFFFFFF)
		{
			fputc((int)tmpchar,tmpout);
		}
	}
	fputc(0x00,tmpout);  //used to find EOF later on
	fclose(tmpout);
	tmpout = fopen("in.tmp","r");
	if(tmpout == NULL)
	{
		fprintf(stderr,"Error opening file\n");
		exit(1);	
	}
	
	
	
	
	

	for(x = 0;x <= 1023 && (fgetc(tmpout)) != 0x00; x++)
	{
		
		fseek(tmpout,-1,SEEK_CUR);

			
		for(;(temp = (fgetc(tmpout))) == 0x20 || temp == 0x09;) ;
		temp = 0;
		fseek(tmpout,-1,SEEK_CUR);
		do
		{
			tmpchar = fgetc(tmpout);
			temp++;
		}while (tmpchar != 0x0A && tmpchar != 0x3B);
		
		fseek(tmpout,-temp,SEEK_CUR);	
		if(temp > 255) 
		{
			printf("Error: Command longer than 255 chars on line %d.\n",x-1);
			exit(1);
		}
		fgets((char *)linemap[x],temp,tmpout);
		if(tmpchar == 0x3B)		//weed out comments
		{
			do
			{
				tmpchar = fgetc(tmpout);
			}while (tmpchar != 0x0D);
		}
		else fseek(tmpout,1,SEEK_CUR);
		//fprintf(stdout,"%s\n",(char *)linemap[x]); 
		
		
	}
	
	assemble(linemap, x);
	
	fcloseall();
	
	return 0;
}

int assemble(char linemap[1024][256], int eof)
{
	int pos, temp, x, line;  //just temporary vars, change to make more sense
	char hold[32][16]; //16 char long device 
	char gotdevice = 0;
	int repeatmask = 0;
	unsigned short int fuse[2] = {0x0FFB,0x0BFF};  //default defined by Parallax SASM, 0 is FUSE, 1 is FUSEX
	unsigned short int id[16];
	unsigned short int optemp, optemp1;  //for instructions with multiple opcodes
	char templine[256];
	for(line = 0; line <= 127;line++)
	{
		for(pos = 0; pos <= 15;pos++)
		{
			opcodes[line][pos] = 0x0FFF;  //initialize all of opcodes
		}
	}
	
	for(line = 0;line <= eof; line++)  //get the device and ID lines all done first so we know how to lay out everything else.
	{
		
		if(!strncmp((char *)linemap[line],"DEVICE",6))  //Go through and see if device line is there, if typo then we will catch it later 
		{													
			printf("Got Device line\n");
			//fuse((char *)linemap[line]);
			for(temp = 0;temp <= 15;temp++)
			{
				for(pos = 0; pos <= 15;pos++)
				{
					hold[temp][pos] = 0x00;
				}
			}
			pos = 6; 	//offset for the word "device"
			for(x = 0; x <= 31 && linemap[line][pos] != 0x00;x++) //copy device settings to arrays
			{
				for(; (int)linemap[line][pos] == 0x20 || (int)linemap[line][pos] == 0x09;pos++) ;  //remove leading spaces and tabs
				for(temp = 0;(int)linemap[line][pos] != 0x2C && (int)linemap[line][pos] != 0x00 && (int)linemap[line][pos] != 0x0D;pos++)
				{
					hold[x][temp] = (int)linemap[line][pos];
					temp++;
				}
				if((int)linemap[line][pos] == 0x2C) pos++;
		
			}
			for(x = 0; x <= 31 && hold[x][0] != 0x00 ;x++)
			{
		
				if(!strncmp((char *)hold[x],"SX48",4) || !strncmp((char *)hold[x],"SX48BD",6) || !strncmp((char *)hold[x],"PINS48",6))
				{
					fprintf(stderr,"SX48/52 Not yet supported\n");
					exit(1);
				}
				if(!strncmp((char *)hold[x],"SX52",4) || !strncmp((char *)hold[x],"SX52BD",6) || !strncmp((char *)hold[x],"PINS52",6))
				{
					fprintf(stderr,"SX48/52 Not yet supported\n");
					exit(1);
				}
		
				if(!strncmp((char *)hold[x],"SX18",4) || !strncmp((char *)hold[x],"SX18AC",6) || !strncmp((char *)hold[x],"PINS18",6))
				{
					fprintf(stdout,"SX18 Device detected\n");
					sx182028 = TRUE;
					gotdevice = x;
				}
				if(!strncmp((char *)hold[x],"SX20",4) || !strncmp((char *)hold[x],"SX20AC",6) || !strncmp((char *)hold[x],"PINS20",6))
				{
					fprintf(stdout,"SX20 Device detected\n");
					sx182028 = TRUE;
					gotdevice = x;
				}
				if(!strncmp((char *)hold[x],"SX28",4) || !strncmp((char *)hold[x],"SX28AC",6) || !strncmp((char *)hold[x],"PINS28",6))
				{
					fprintf(stdout,"SX28 Device detected\n");
					fuse[1] = (fuse[1] + 0x400);         //PINS28 is bit 10 of FUSEX
					sx182028 = TRUE;
					gotdevice = x;
				}
		
			}
			if(sx182028)
			{
				for(pos = 0; pos <= 31 && hold[pos][0] != 0x00;pos++)
				{
					if(hold[pos] == hold[(int)gotdevice]) ;
					else if(!strncmp((char *)hold[pos],"BANKS1",6))   //bit 0 and 1 of FUSEX
					{
						if((repeatmask & 0x01))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BANKS setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x03))  //make sure its BANKS8 already
						{
							//fuse[1] = (fuse[1] - 0x03);
							repeatmask = repeatmask^0x01;
						}
						else 
						{
							fprintf(stderr,"Error setting BANKS1\n");
							exit(1);
						}
					}
					else if(!strncmp((char *)hold[pos],"BANKS2",6))   //bit 0 and 1 of FUSEX
					{
						if((repeatmask & 0x01))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BANKS setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x03))  //make sure its BANKS8 already
						{
							//fuse[1] = (fuse[1] - 0x03);
							repeatmask = repeatmask^0x01;
						}
						else 
						{
							fprintf(stderr,"Error setting BANKS2\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"BANKS4",6))   //bit 0 and 1 of FUSEX
					{
						if((repeatmask & 0x01))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BANKS setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x03))   //make sure its BANKS8 already
						{
							//fuse[1] = (fuse[1] - 0x03);
							repeatmask = repeatmask^0x01;
						}
						else 
						{
							fprintf(stderr,"Error setting BANKS4\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"BANKS8",6))   //bit 0 and 1 of FUSEX
					{
						if((repeatmask & 0x01))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BANKS setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x03))   //make sure its BANKS8 already
						{
							//fuse[1] = (fuse[1] - 0x03);
							repeatmask = repeatmask^0x01;
						}
						else 
						{
							fprintf(stderr,"Error setting BANKS8\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSCHS3",6))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x01);  //Set OSCSH3
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCHS3\n");
							exit(1);
						}
				
				
					}
					else if(!strncmp((char *)hold[pos],"OSCHS2",6))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{	
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x02);  //Set OSCSH2
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCHS2\n");
							exit(1);
						}
					}
					else if(!strncmp((char *)hold[pos],"OSCHS1",6))   //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x03);  //Set OSCSH1
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCHS1\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSCXT2",6))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x20);  //Set OSCXT2
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCXT2\n");  
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSCXT1",6))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x21);  //Set OSCXT1
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCXT1\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSCLP2",6))   //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x22);  //Set OSCLP2
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCLP2\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"OSCLP1",6))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							fuse[0] = (fuse[0] - 0x23);  //Set OSCLP1
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCLP1\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSCRC",5))  //bit 0 1 5 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x23))  //make sure it is OSCRC already
						{
							repeatmask = repeatmask^0x02;
						}
						else 
						{
							fprintf(stderr,"Error setting OSCRC\n");
							exit(1);
						}
					}
					else if(!strncmp((char *)hold[pos],"OSC4MHZ",7))  //bit 5 6 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((repeatmask & 0x04))  //make sure IFBD isnt used
						{
							fprintf(stderr,"Error: Line %d: Cannot use both IFBD and internal OSC\n",line + 1);
							exit(1);
						}
						else if((fuse[0] & 0x23))  //make sure it is OSCRC already and IFBD is not set yet
						{
							fuse[0] = ((fuse[0] - 0x60) - 0x80);  //Set OSC4MHZ and set internal oscilator
							repeatmask = repeatmask^0x02;
							repeatmask = repeatmask^0x1000;
						}
						else 
						{
							fprintf(stderr,"Error setting OSC4MHZ\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"OSC1MHZ",7))  //bit 5 6 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((repeatmask & 0x04))  //make sure IFBD isnt used
						{
							fprintf(stderr,"Error: Line %d: Cannot use both IFBD and internal OSC\n",line + 1);
							exit(1);
						}
						else if((fuse[0] & 0x23))  //make sure it is OSCRC already enabled and IFBD is not set yet
						{
							fuse[0] = ((fuse[0] - 0x40) - 0x80);  //Set OSC1MHZ and set internal oscilator
							repeatmask = repeatmask^0x02;
							repeatmask = repeatmask^0x1000;
						}
						else 
						{
							fprintf(stderr,"Error setting OSC1MHZ\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSC128KHZ",9))  //bit 5 6 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((repeatmask & 0x04))  //make sure IFBD isnt used
						{
							fprintf(stderr,"Error: Line %d: Cannot use both IFBD and internal OSC\n",line + 1);
							exit(1);
						}
						else if((fuse[0] & 0x23))  //make sure it is OSCRC already and IFBD is not enabled
						{
							fuse[0] = (((fuse[0] - 0x20) + 0x40) - 0x80);  //Set OSC128KHZ and set internal oscilator
							repeatmask = repeatmask^0x02;
							repeatmask = repeatmask^0x1000;
						}
						else 
						{
							fprintf(stderr,"Error setting OSC128KHZ\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"OSC32KHZ",8))  //bit 5 6 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x02))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of OSC setting\n",line + 1);
							exit(1);
						}
						if((repeatmask & 0x04)) //make sure IFBD isnt used
						{
							fprintf(stderr,"Error: Line %d: Cannot use both IFBD and internal OSC\n",line + 1);
							exit(1);
						}
						else if((fuse[0] & 0x23))  //make sure it is OSCRC already and IFBD is not yet set
						{
							fuse[0] = ((fuse[0] + 0x60) - 0x80);  //Set OSC32KHZ and set internal oscilator
							repeatmask = repeatmask^0x02;
							repeatmask = repeatmask^0x1000;
						}
						else 
						{
							fprintf(stderr,"Error setting OSC32KHZ\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"IFBD",4))  //bit 6 of FUSE, IRC is bit 7 of FUSE
					{
						if((repeatmask & 0x04))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of IFBD setting\n",line + 1);
							exit(1);
						}
						if((repeatmask & 0x1000))  //make sure no internal OSC device is declared
						{
							fprintf(stderr,"Error: Line %d: Cannot use both IFBD and internal OSC\n",line + 1);
							exit(1);
						}
						else if((fuse[0] & 0x80)) //make sure it is IRC is zero
						{
							fuse[0] = (fuse[0] - 0x40);  //Set IFBD
							repeatmask = repeatmask^0x04;
						}
						else 
						{
							fprintf(stderr,"Error setting IFBD, Not correctly implemented yet\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"BOR42",5))  //bit 4 5 of FUSEX
					{
						if((repeatmask & 0x10))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BOR setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x30))  //make sure it is BOROFF already
						{
							fuse[1] = (fuse[1] - 0x30);  //Set BOR42
							repeatmask = repeatmask^0x10;
						}
						else 
						{
							fprintf(stderr,"Error setting BOR42\n");
							exit(1);
						}
					
					}
					else if(!strncmp((char *)hold[pos],"BOR26",5))  //bit 4 5 of FUSEX
					{
						if((repeatmask & 0x10))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BOR setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x30))  //make sure it is BOROFF already
						{
							fuse[1] = (fuse[1] - 0x20);  //Set BOR26
							repeatmask = repeatmask^0x10;
						}
						else
						{
							fprintf(stderr,"Error setting BOR26\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"BOR22",5))  //bit 4 5 of FUSEX
					{
						if((repeatmask & 0x10))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BOR setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x30))  //make sure it is BOROFF already
						{
							fuse[1] = (fuse[1] - 0x10);  //Set BOR26
							repeatmask = repeatmask^0x10;
						}
						else 
						{
							fprintf(stderr,"Error setting BOR22\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"BOROFF",6))  //bit 4 5 of FUSEX
					{
						if((repeatmask & 0x10))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of BOR setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x30))  //make sure it is BOROFF already
						{
							repeatmask = repeatmask^0x10;
						}
						else 
						{
							fprintf(stderr,"Error setting BOROFF\n");
							exit(1);
						}

					}
					else if(!strncmp((char *)hold[pos],"TURBO",5))  //bit 11 of FUSE
					{
						if((repeatmask & 0x20))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of TURBO setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x800))  //make sure it is TURBO on
						{
							fuse[0] = (fuse[0] - 0x800);
							repeatmask = repeatmask^0x20;
						}
						else 
						{
							fprintf(stderr,"Error setting TURBO\n");
							exit(1);
						}
						
					}
					else if(!strncmp((char *)hold[pos],"OPTIONX",7) || !strncmp((char *)hold[pos],"STACKX",6))  //bit 7 of FUSEX
					{
						if((repeatmask & 0x40))
						{
							//fprintf(stderr,"Error: Line %d: Redefinition of OPTIONX/STACKX setting\n",line + 1);
							//exit(1);
							break;
						}
						if((fuse[1] & 0x80))  //make sure OPTIONX/STACKX is disabled
						{
							fuse[1] = (fuse[1] - 0x80);
							repeatmask = repeatmask^0x40;
						}
						else 
						{
							fprintf(stderr,"Error setting OPTIONX/STACKX\n");
							exit(1);
						}
				
				
					}
					else if(!strncmp((char *)hold[pos],"CARRYX",6))  //bit 6 of FUSEX
					{
						if((repeatmask & 0x80))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of CARRYX setting\n",line + 1);
							exit(1);
						}
						if((fuse[1] & 0x40))  //make sure CARRYX is disabled
						{
							fuse[1] = (fuse[1] - 0x40);
							repeatmask = repeatmask^0x80;
						}
						else 
						{
							fprintf(stderr,"Error setting CARRYX\n");
							exit(1);
						}
				
				
					}
					else if(!strncmp((char *)hold[pos],"SYNC",4))  //bit 10 of FUSE
					{
						if((repeatmask & 0x200))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of SYNC setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x400))  //make sure SYNC is disabled
						{
							fuse[0] = (fuse[0] - 0x400);  
							repeatmask = repeatmask^0x200;
						}
						else 
						{
							fprintf(stderr,"Error setting SYNC\n");
							exit(1);
						}
				
					}
					else if(!strncmp((char *)hold[pos],"WATCHDOG",8))  //bit 2 of FUSE
					{
						if((repeatmask & 0x400))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of WATCHDOG setting\n",line + 1);
							exit(1);
						}
						if(!(fuse[0] & 0x04))  //make sure WATCHDOG id disabled
						{
							fuse[0] = (fuse[0] + 0x04);
							repeatmask = repeatmask^0x400;
						}
						else 
						{
							fprintf(stderr,"Error setting WATCHDOG\n");
							exit(1);
						}
					}
					else if(!strncmp((char *)hold[pos],"PROTECT",7))  //bit 3 of FUSE
					{
						if((repeatmask & 0x800))
						{
							fprintf(stderr,"Error: Line %d: Redefinition of PROTECT setting\n",line + 1);
							exit(1);
						}
						if((fuse[0] & 0x08))  //make sure PROTECT is disabled
						{
							fuse[0] = (fuse[0] - 0x08);
							repeatmask = repeatmask^0x800;
						}
						else 
						{
							fprintf(stderr,"Error setting PROTECT\n");
							exit(1);
						}
					}
					else if(!strncmp((char *)hold[pos],"SX18",4) || !strncmp((char *)hold[pos],"SX18AC",6) || 
					!strncmp((char *)hold[pos],"PINS18",6) || !strncmp((char *)hold[pos],"SX20",4) || 
					!strncmp((char *)hold[pos],"SX20AC",6) || !strncmp((char *)hold[pos],"PINS20",6) || 
					!strncmp((char *)hold[pos],"SX28",4) || !strncmp((char *)hold[pos],"SX28AC",6) || 
					!strncmp((char *)hold[pos],"PINS28",6) || !strncmp((char *)hold[pos],"SX48",4) || 
					!strncmp((char *)hold[pos],"SX48BD",6) || !strncmp((char *)hold[pos],"PINS48",6) || 
					!strncmp((char *)hold[pos],"SX52",4) || !strncmp((char *)hold[pos],"SX52BD",6) || 
					!strncmp((char *)hold[pos],"PINS52",6))
					{
						fprintf(stderr,"Error: Line %d: Redeclaration of SX device type: \"%s\"\n",line,hold[pos]);
						exit(1);
					}
					else 
					{
						fprintf(stderr,"Error: Line %d: Unknown Device Directive \"%s\"\n",line,hold[pos]);
						//add something to an error variable so if any errors display them all, then exit out later
					}
				}
			}
		}
		else if(!strncmp((char *)linemap[line],"ID",2))  //same as above with id 
		{													
			printf("Got id line, not doing anything with it right now\n");
			//id((char *)linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"RESET",5))
		{
			//if(repeatmask & 0x1000)
			//{
			//	fprintf(stderr,"Error: Line %d: Redefinition of RESET\n",line + 1);
			//	exit(1);
			//}
			for(pos = 5; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
			optemp = jmp(linemap[line],pos,line,0);
			/*optemp = literalize(linemap[line],pos,line);
			optemp <<= 7;
			optemp >>= 7;
			optemp += 0x0A00;*/
			opcodes[127][15] = optemp;
				
		}
		else if(!strncmp((char *)linemap[line],"IRC_CAL",7))
		{
			if(sx182028)
			{
				for(pos = 8; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //ignore spaces after 
				if(!strncmp((char *)linemap[line] + pos,"IRC_SLOW",8))  //IRCTRIM0-1-2??
				{
					if((repeatmask & 0x08))
					{
						fprintf(stderr,"Error: Line %d: Redefinition of IRC setting\n",line + 1);
						exit(1);
					}
					if((fuse[1] & 0x0B00)) //make sure we have default
					{
						fuse[1] = (fuse[1] - 0xB00);
					}	
					repeatmask = repeatmask^0x08;
				
				}
				else if(!strncmp((char *)linemap[line] + pos,"IRC_4MHZ",8))  //IRCTRIM0-1-2??
				{	
					if((repeatmask & 0x08))
					{
						fprintf(stderr,"Error: Line %d: Redefinition of IRC setting\n",line + 1);
						exit(1);
					}
					if((fuse[1] & 0x0B00)) //make sure we have default
					{
						//fuse[1] = (fuse[1] - 0xB00);
						fprintf(stderr,"Warning: Line %d: IRC_4MHZ calibration not implemented, defaulting to IRC_FAST\n",line);
					}	
					repeatmask = repeatmask^0x08;
					
				}
				else if(!strncmp((char *)linemap[line] + pos,"IRC_FAST",8))  //IRCTRIM0-1-2??
				{
					if((repeatmask & 0x08))
					{
						fprintf(stderr,"Error: Line %d: Redefinition of IRC setting\n",line + 1);
						exit(1);
					}
					if((fuse[1] & 0x0B00)) //make sure we have default
					{
						//fuse[1] = (fuse[1] - 0xB00);  do nothing cuz fast is default
					}	
					repeatmask = repeatmask^0x08;
						
				}
			}
			else 
			{
				printf("Generic Error\n");
				exit(1);
			}
		}
	}	
	for(line = 0;line < eof; line++)
	{
		if(linemap[line][0] == 0x0A || linemap[line][0] == 0x0D || linemap[line][0] == 0x00 ) ;//|| linemap[line][0] == 0x3B) ;
		else if(!strncmp((char *)linemap[line],"DEVICE",6)) ;
		else if(!strncmp((char *)linemap[line],"ID",2)) ;
		else if(!strncmp((char *)linemap[line],"RESET",5)) ;
		else if(!strncmp((char *)linemap[line],"FREQ",4)) 
		{
			fprintf(stderr,"Warning: Line %d: FREQ directive is ignored\n",line + 1);
		}
		else if(!strncmp((char *)linemap[line],"IRC_CAL",7));
		else if(!strncmp((char *)linemap[line],"ORG",3))
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after ORG and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			opline = 0;
			for(;optemp > 16;)
			{
				optemp = optemp - 16;
				opline++;
			}
			oppos = optemp;
			
		}			
		else if(!strncmp((char *)linemap[line],"ADDB",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after ADDB and remove spaces/tabs
			optemp1 = literalize(linemap[line],pos,line);  //save fr
			for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				//printf("%c\n",linemap[line][pos]);
				optemp = dobit(linemap[line],pos,line);
				if(linemap[line][pos] != '/') optemp += 0x0600;  //"snb"
				else if(linemap[line][pos] == '/') optemp += 0x0700;  //"sb"
				else bad_expr(line,linemap[line]);
				write_array(optemp,line);
				optemp1 <<= 11;
				optemp1 >>= 11;
				optemp1 += 0x02A0; //"inc fr"
				write_array(optemp1,line);
			}else bad_expr(line,linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"ADD",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after ADD and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //this mean we will have "add w,fr" opcode 0x01C0
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ;  //ignore spaces
					if(linemap[line][pos] == '#') bad_expr(line,linemap[line]);  //make sure its not a literal
					else 
					{					
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						//optemp &= ~(0x1E0);    //zero out that opcode area leaving the register value  figure out why it wont work
						optemp += 0x01C0;  //opcode for "mov w,fr"  
					}
				}
				else bad_expr(line,linemap[line]);
			}
			else  //could be add fr,#lit or add fr1,fr2 or add fr,w
			{
				optemp1 = literalize(linemap[line],pos,line);  //store fr1
				for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
				if(linemap[line][pos] == ',')
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == '#')  //means it is add fr,#lit
					{							  //expands to mov w,#lit add fr,w
						optemp = mov_w_lit(linemap[line],pos,line,0);  //let function literalize
						write_array(optemp, line);
						optemp = add_fr_w(linemap[line],pos,line,optemp1);  //pass fr1 to it
					}
					else if((linemap[line][pos] == 'W'))  //means we have add fr,w
					{
						optemp = add_fr_w(linemap[line],pos,line,optemp1);
					}
					else    //means we have "add fr1,fr2"
					{		//expands to "mov w,fr2" "add fr1,w"
						
						//we are currently at fr2
						optemp = mov_w_fr(linemap[line],pos,line,0); //let function literalize
						write_array(optemp, line);
						optemp = add_fr_w(linemap[line],pos,line,optemp1); //pass fr1 to it
					}
				}
				else bad_expr(line,linemap[line]);
			}
			write_array(optemp, line);		
		}			
		else if(!strncmp((char *)linemap[line],"AND",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after AND and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //means we can have "and w,#lit" or "and w,fr"
			{
				for(; linemap[line][pos] != ',';pos++) ;
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp = literalize(linemap[line],pos,line);
				if(linemap[line][pos] == '#') //confirms "and w,#lit"
				{
					optemp <<= 8;
					optemp >>= 8;
					optemp += 0x0E00;
				}
				else  //we probably have "and w,fr" let literalize() do our error handling
				{
					optemp <<= 11;
					optemp >>= 11;
					optemp += 0x0140;
				}
			}
			else  //we could have "and fr,w" "and,fr,#lit" "and fr1,fr2"
			{
				optemp1 = literalize(linemap[line],pos,line);  //save fr1
				for(; linemap[line][pos] != ',';pos++) ;  //we have fr1 saved, now get the second argument
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == 'W')  //confirms "and fr,w"
				{
					optemp = and_fr_w(linemap[line],pos,line,optemp1);  //pass fr1 to it so we dont have to redo literalize()
				}
				else if(linemap[line][pos] == '#') //confirms "and fr,#lit"  breaks into "mov w,#lit" "and fr,w"
				{
					//we now have the fr as optemp1 and lit as optemp
					optemp = mov_w_lit(linemap[line],pos,line,0);  //will take care of getting value 
					write_array(optemp, line);
					optemp = and_fr_w(linemap[line],pos,line,optemp1); //pass fr to this
				}
				else //we should have "and fr1,fr2"
				{	//expands to "mov w,fr2" "and fr1,w"
					optemp = mov_w_fr(linemap[line],pos,line,0);  //let function get literal
					write_array(optemp, line);
					optemp = and_fr_w(linemap[line],pos,line,optemp1);  //pass fr1 to finction
				}
					
			}
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"BANK",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after BANK and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp >>= 4;
			optemp /= 2;
			optemp <<= 13;
			optemp >>= 13;
			optemp += 0x0018;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"CALL",4)) //working
		{
			fprintf(stderr,"Warning: Line %d: CALL command not regulated for proper use, be careful\n",line);
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after CALL and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 8;
			optemp >>= 8;
			optemp += 0x0900;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"CJAE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,1);
		}
		else if(!strncmp((char *)linemap[line],"CJA",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,0);  //0 sine it is cja
		}
		else if(!strncmp((char *)linemap[line],"CJBE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,3);  //3 is cjbe
		}
		else if(!strncmp((char *)linemap[line],"CJB",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,2);
		}
		else if(!strncmp((char *)linemap[line],"CJE",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,4);
		}
		else if(!strncmp((char *)linemap[line],"CJNE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,5);
		}
		else if(!strncmp((char *)linemap[line],"CLC",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);
			//for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after CLC and remove spaces/tabs	
			optemp = 0x0403;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"CLRB",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0400;
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"CLR",3)) //works
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after CLR and remove spaces/tabs
			if(linemap[line][pos] == 'W')
			{
				optemp = 0x0040;
				write_array(optemp, line);
			}
			else if(linemap[line][pos] == '!')  //we probably have clr !WDT
			{
				if((linemap[line][pos + 1] == 'W') && (linemap[line][pos + 2] == 'D') && (linemap[line][pos + 3] == 'T'))
				{ //yeah, clr !WDT
					optemp = 0x0004;
					write_array(optemp, line);
				}
				else bad_expr(line,linemap[line]);
			}
			else  //it should be clr fr
			{
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 11;
				optemp >>= 11;
				optemp += 0x0060;
				write_array(optemp, line);
			}
		}
		else if(!strncmp((char *)linemap[line],"CLZ",3)) //works
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);
			optemp = 0x0443;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"CSAE",4)) //woroking
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,7);  //7 is csae
		}
		else if(!strncmp((char *)linemap[line],"CSA",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,6);  //6 is csa
		}
		else if(!strncmp((char *)linemap[line],"CSBE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,9);  //9 is csbe
		}
		else if(!strncmp((char *)linemap[line],"CSB",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,8);  //8 is csb
		}
		else if(!strncmp((char *)linemap[line],"CSNE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,11);  //11 is csne
		}
		else if(!strncmp((char *)linemap[line],"CSE",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			cjs(linemap[line],pos,line,10);  //10 is cse
		}
		else if(!strncmp((char *)linemap[line],"DECSZ",5)) //works
		{
			for(pos = 5; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after DECSZ and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x02E0;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"DEC",3)) //works
		{
			
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after DEC and remove spaces/tabs

			
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x00E0;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"DJNZ",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after DJNZ and remove spaces/tabs
			//we just have to assume it is fr,addr9
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x02E0;
			write_array(optemp, line);
			for(; linemap[line][pos] != ',' && linemap[line][pos] != 0x0D;pos++) ; 
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp = jmp(linemap[line],pos,line,0);
				/*optemp = literalize(linemap[line],pos,line);
				optemp <<= 7; 
				optemp >>= 7;
				optemp += 0x0A00;*/
				write_array(optemp, line);
			}
			else
			{
				fprintf(stderr,"Weird Error\n");
				exit(1);
			}
		}			
		else if(!strncmp((char *)linemap[line],"IJNZ",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after IJNZ and remove spaces/tabs
			//we just have to assume it is fr,addr9
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x03E0;
			write_array(optemp, line);
			for(; linemap[line][pos] != ',' && linemap[line][pos] != 0x0D;pos++) ; 
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp = jmp(linemap[line],pos,line,0);
				/*optemp = literalize(linemap[line],pos,line);
				optemp <<= 7; 
				optemp >>= 7;
				optemp += 0x0A00;*/
				write_array(optemp, line);
			}
			else
			{
				fprintf(stderr,"Weird Error\n");
				exit(1);
			}
		}
		else if(!strncmp((char *)linemap[line],"INCSZ",5)) //works
		{
			if((linemap[line][5] != 0x20) && (linemap[line][5] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 5; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after INCSZ and remove spaces/tabs
			
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x03E0;
			write_array(optemp, line);
		}		
		else if(!strncmp((char *)linemap[line],"INC",3)) //works
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after INC and remove spaces/tabs
			
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x02A0;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"IREAD",5)) //works
		{
			if((linemap[line][5] != 0x20) && (linemap[line][5] != 0x09) && (linemap[line][5] != 0x0D)) bad_expr(line,linemap[line]);  //fix to make sure nothing else is on the line
			optemp = 0x0041;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"JB",2)) //working
		{
			//breaks into "snb op.bit" "jmp addr9"
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 2; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0600; //"snb"
			write_array(optemp,line);
			for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 8;
				optemp >>= 8;
				optemp += 0x0A00; //"jmp"
				write_array(optemp,line);
			}else bad_expr(line,linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"JC",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);
			j(linemap[line],pos,line,1); //1 is jc
		}
		else if(!strncmp((char *)linemap[line],"JMP",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09 || linemap[line][pos] == 0x0D;pos++) ; //start off reading after JMP and remove spaces/tabs
			optemp = jmp(linemap[line],pos,line,0);  //let function get addr9
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"JNB",3)) //working
		{
			//breaks into "snb op.bit" "jmp addr9"
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0700; //"sb"
			write_array(optemp,line);
			for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 8;
				optemp >>= 8;
				optemp += 0x0A00; //"jmp"
				write_array(optemp,line);
			}else bad_expr(line,linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"JNC",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			j(linemap[line],pos,line,3); //3 is jnc
		}
		else if(!strncmp((char *)linemap[line],"JNZ",3)) //works
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
			j(linemap[line],pos,line,4); //4 is jnz
		}
		else if(!strncmp((char *)linemap[line],"JZ",2)) //works
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);
			j(linemap[line],pos,line,5); //5 is jz
		}
		else if(!strncmp((char *)linemap[line],"MODE",4)) //works    change this for sx48/52 for the full mode move 
		{
			if(sx182028)
			{
				for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after MODE and remove spaces/tabs
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 12;
				optemp >>= 12;				
				optemp += 0x0050;
				write_array(optemp, line);
			}
				
		}
		else if(!strncmp((char *)linemap[line],"MOVB",4)) //working
		{  //breaks into "sb op.bit2" "clrb op.bit1" "snb op.bit2" "setb op.bit1"
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SUBB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				optemp1 = dobit(linemap[line],pos,line);
			}else bad_expr(line,linemap[line]);
			
			if(linemap[line][pos] != '/') write_array(optemp1 + 0x0700,line); //"sb op.bit2"
			else if(linemap[line][pos] == '/') write_array(optemp1 + 0x0600,line); //"snb op.bit2"
			else bad_expr(line,linemap[line]);
			write_array(optemp + 0x0400,line); //"clrb op.bit1"
			if(linemap[line][pos] != '/') write_array(optemp1 + 0x0600,line); //"snb op.bit2"
			else if(linemap[line][pos] == '/') write_array(optemp1 + 0x0700,line); //"sb op.bit2"
			else bad_expr(line,linemap[line]);
			write_array(optemp + 0x0500,line);
		}
		else if(!strncmp((char *)linemap[line],"MOVSZ",5)) //working
		{
			if((linemap[line][5] != 0x20) && (linemap[line][5] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 5; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after ADD and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //were on the right track
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces	
					if((linemap[line][pos] == '-'))  //confirmed "movsz w,--fr"
					{

						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] != '-')  bad_expr(line,linemap[line]);  //we only have "movsz w, -fr" which is wrong
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x02C0;
					}
					else if((linemap[line][pos] == '+'))  //confirmed "movsz w,++fr"
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] != '+')  bad_expr(line,linemap[line]);  //we only have "movsz w, +fr" which is wrong
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x03C0;
					}	
					else bad_expr(line,linemap[line]);				
				}
				else bad_expr(line,linemap[line]);
			}
			else
			{
				fprintf(stderr,"Generic Error\n");
				exit(1);
			}
			write_array(optemp, line);
		}	
		else if(!strncmp((char *)linemap[line],"MOV",3)) //working
		{
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after ADD and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //this mean we will have "mov w,something"
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == '#')   //we have mov "w,lit" opcode 0x0C00
					{
						optemp = mov_w_lit(linemap[line],pos,line,0);
					}
					else if(linemap[line][pos] == '/')  //confirmed "mov w,/fr"
					{
						pos++;  
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line); //literalize will take care of the rest
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x0240;
					}
					else if((linemap[line][pos] == '+')) //confirmed "mov w,++fr"
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] != '+')  bad_expr(line,linemap[line]);  //we only have "mov w, -fr" which is wrong
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x0280;
					}
					else if((linemap[line][pos] == '-'))  //confirmed "mov w,--fr"
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] != '-')  bad_expr(line,linemap[line]);  //we only have "mov w, -fr" which is wrong
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x00C0;
					}
					else if((linemap[line][pos] == '>'))  //confirmed "mov w,>>fr"
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] != '>')  bad_expr(line,linemap[line]);  //we only have "mov w, -fr" which is wrong
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x0300;
					}
					else if((linemap[line][pos] == '<'))  //could be "mov w,<<fr" or "mov w,<>fr"
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] == '<')  //confirmed "mov w,<<fr"
						{
							pos++;
							for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
							optemp = literalize(linemap[line],pos,line);
							optemp <<= 11;
							optemp >>= 11;
							optemp += 0x0340;
						}
						else if(linemap[line][pos] == '>')  //confirmed "mov w,<>fr"
						{
							pos++;
							for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
							optemp = literalize(linemap[line],pos,line);
							optemp <<= 11;
							optemp >>= 11;
							optemp += 0x0380;
						}
					}
					else if(linemap[line][pos] == 'M')  //confirmed "mov w,m"
					{
						optemp = 0x0042;
					}
					else //we can assume it is "mov w,fr" or "mov w,fr-W"
					{
						for(temp = 0; (linemap[line][pos + temp] != '-') && (linemap[line][pos + temp] != 0x0D); temp++);
						//printf("%X\n",linemap[line][pos + temp]);
						//printf("%c\n", linemap[line][pos + temp]);
						if(linemap[line][pos + temp] == 0x0D) //confirms "mov w,fr"
						{
							optemp = mov_w_fr(linemap[line],pos,line,0);  //we let function literalize
						}
						else if(linemap[line][pos + temp] == '-') //we should have "mov w,fr-w" 
						{
							for(x = 0;(linemap[line][pos + temp + x] != 'W') && (linemap[line][pos + temp + x] != 0x0D) ;x++) ; //make sure there is actually a W after -
							if(linemap[line][pos + temp + x] != 'W') bad_expr(line,linemap[line]);
							strncpy(templine,linemap[line],(pos + temp));  //copy up to "-" to another string
							optemp = literalize(templine,pos,line);  //since direct copy, pos will hold correct position
							optemp <<= 11;
							optemp >>= 11;
							optemp += 0x0080;
						}
							
					}
				}
			}
			else if(linemap[line][pos] == 'M')  //means we have "mov m,something"
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == 'W')  //confirms "mov m,w"
					{
						optemp = 0x0043;
					}
					else if(linemap[line][pos] == '#')  //confirms "mov w,#lit"
					{
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 12;
						optemp >>= 12;
						optemp += 0x0050;
					}
					else //we can assume "mov m,fr"
					{   //breaks down to "mov w,fr"  "mov m,w"
						optemp = mov_w_fr(linemap[line],pos,line,0); //let function literalize
						write_array(optemp, line);
						optemp = 0x0043;
					}
				}
				else bad_expr(line,linemap[line]);
			}
			else if(linemap[line][pos] == '!')  //means we have "mov !port,something" or "mov !OPTION,something"
			{
				if(!strncmp((char *)linemap[line] + pos,"!OPTION",7)) //we have "mov !option,something"
				{
					pos += 7;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == ',') //make sure we have a ,
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] == 'W')  //confirms "mov !option,w"
						{
							optemp = 0x0002;
						}
						else if(linemap[line][pos] == '#')  //confirms "mov !option,#lit"
						{								   //break into "mov w,#lit" "mov !option,w"
							optemp = mov_w_lit(linemap[line],pos,line,0);
							write_array(optemp, line);
							optemp = 0x0002;
						}
						else //we can assume "mov !option,fr"
						{   //breaks into "mov w,fr" "mov !option,w"
							optemp = mov_w_fr(linemap[line],pos,line,0);
							write_array(optemp, line);
							optemp = 0x0002;
						}
					}
					else bad_expr(line,linemap[line]);
				}
				else //assume we have "mov !port,something" let literalize check
				{
					pos++;
					optemp1 = literalize(linemap[line],pos,line);  //get port number
					for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D) ;pos++) ; //go to the end
					if(linemap[line][pos] == ',') //make sure we have a , 
					{
						pos++;
						for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
						if(linemap[line][pos] == 'W')  //confirms "mov !port,w"
						{
							optemp = optemp1;
						}
						else if(linemap[line][pos] == '#')  //confirms "mov !port,#lit"
						{
							optemp = mov_w_lit(linemap[line],pos,line,0);
							write_array(optemp, line);
							optemp = optemp1;
						}
						else //we can assume "mov !port,fr"
						{
							optemp = mov_w_fr(linemap[line],pos,line,0);
							write_array(optemp, line);
							optemp = optemp1;
						}							
					}
					else bad_expr(line,linemap[line]);
				}
			}
			else //we can assume it is "mov fr,something"
			{
				optemp1 = literalize(linemap[line],pos,line);  //store fr1
				for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
				if(linemap[line][pos] == 0x0D) bad_expr(line,linemap[line]);
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == 'W')  //confirms "mov fr,w"
					{
						optemp = mov_fr_w(linemap[line],pos,line,optemp1);  //we alread have fr1
					}
					else if(linemap[line][pos] == 'M')  //confirms "mov fr,m"
					{								   //breaks into "mov w,m" "mov w,fr"
						optemp = 0x0042;  //opcode for "mov w,m"
						write_array(optemp, line);
						optemp = mov_fr_w(linemap[line],pos,line,optemp1); //we already have fr1
					}
					else if(linemap[line][pos] == '#')  //confirms "mov fr,#lit"
					{                                  //breaks into "mov w,#lit" "mov fr,w"
						optemp = mov_w_lit(linemap[line],pos,line,0);  //let funtction literalize
						write_array(optemp, line);
						optemp = mov_fr_w(linemap[line],pos,line,optemp1); //we have fr1
					}
					else //assume we have "mov fr1,fr2"
					{   //breaks into "mov w,fr2" "mov fr1,w"
						optemp = mov_w_fr(linemap[line],pos,line,0); //let the function get fr2
						write_array(optemp, line);
						optemp = mov_fr_w(linemap[line],pos,line,optemp1);  //we have fr1 already
					}
				}
				else bad_expr(line,linemap[line]);
			}
				
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"NOP",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0000;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"NOT",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]); 
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after NOT and remove spaces/tabs
			if(linemap[line][pos] == 'W')
			{
				optemp = 0x0FFF;
			}
			else
			{
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 11;
				optemp >>= 11;
				optemp += 0x0260;
			}
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"OR",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 2; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after OR and remove spaces/tabs
			if(linemap[line][pos] == 'W') //could be "or w,fr" "or w,#lit"
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')  //the next operand will be , 
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == '#')  //confirms "or w,#lit"
					{
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 8;
						optemp >>= 8;
						optemp += 0x0D00;
					}
					else //we can assume it is "or w,fr"
					{
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x0100;
					}
				}
				else bad_expr(line,linemap[line]);
			}
			else //we can assume it is "or fr1,w" "or fr,#lit" "or fr1,fr2"
			{
				optemp1 = literalize(linemap[line],pos,line);  //save fr1
				for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
				if(linemap[line][pos] == 0x0D) bad_expr(line,linemap[line]);
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == 'W') //confirms "or fr,w"
				{
					optemp = or_fr_w(linemap[line],pos,line,optemp1); //we already have optemp
				}
				else if(linemap[line][pos] == '#') //confirms "or fr,#lit"
				{								  //breaks into "mov w,#lit" "or fr,w"
					optemp = mov_w_lit(linemap[line],pos,line,0);  //let function get lit
					write_array(optemp, line);
					optemp = or_fr_w(linemap[line],pos,line,optemp1);  //we have fr1
				}
				else //we can assume "or fr1,fr2"
				{   //breaks into "mov w,fr2" "or fr,w"
					optemp = mov_w_fr(linemap[line],pos,line,0);  //let function get fr2
					write_array(optemp, line);
					optemp = or_fr_w(linemap[line],pos,line,optemp1);  //we already have fr1
				}
			}
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"PAGE",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after BANK and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp >>= 8;
			optemp /= 2;
			optemp <<= 13;
			optemp >>= 13;
			optemp += 0x0010;
			write_array(optemp, line);
		}	
		else if(!strncmp((char *)linemap[line],"RETIW",5)) //working
		{
			if((linemap[line][5] != 0x20) && (linemap[line][5] != 0x09) && (linemap[line][5] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x000F;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"RETI",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09) && (linemap[line][4] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x000E;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"RETP",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09) && (linemap[line][4] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x000D;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"RETW",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after RETW and remove spaces/tabs
			for(;linemap[line][pos] != 0x0D && linemap[line][pos] != 0x0A && linemap[line][pos] != 0x0;)
			{
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 8;
				optemp >>= 8;
				optemp += 0x0800;
				write_array(optemp,line);
				pos++;
				for(;linemap[line][pos] == ',' || linemap[line][pos] == 0x09 || linemap[line][pos] == 0x20 /*&& linemap[line][pos] != 0x0D*/;pos++);
			}
			
		}
		else if(!strncmp((char *)linemap[line],"RET",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x000C;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"RL",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 2; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after RL and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x0360;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"RR",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 2; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after RR and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x0320;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SB",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 2; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0700;
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"SC",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09) && (linemap[line][2] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0703;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SETB",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0500;
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"SKIP",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09) && (linemap[line][4] != 0x0D)) bad_expr(line,linemap[line]);  
			if(!(oppos % 2)) //means next instruction will be odd
			{
				optemp = 0x0702;
			}
			else optemp = 0x0602;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SLEEP",5)) //working
		{
			if((linemap[line][5] != 0x20) && (linemap[line][5] != 0x09) && (linemap[line][5] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0003;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SNB",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SB and remove spaces/tabs
			optemp = dobit(linemap[line],pos,line);
			optemp += 0x0600;
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"SNC",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0603;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SNZ",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0643;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"STC",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0503;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"STZ",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09) && (linemap[line][3] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0543;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"SUBB",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SUBB and remove spaces/tabs
			optemp1 = literalize(linemap[line],pos,line);  //save fr
			for(; (linemap[line][pos] != ',') && (linemap[line][pos] != 0x0D);pos++) ;
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				//printf("%c\n",linemap[line][pos]);
				optemp = dobit(linemap[line],pos,line);
				if(linemap[line][pos] != '/') optemp += 0x0600;  //"snb"
				else if(linemap[line][pos] == '/') optemp += 0x0700;  //"sb"
				else bad_expr(line,linemap[line]);
				write_array(optemp,line);
				optemp1 <<= 11;
				optemp1 >>= 11;
				optemp1 += 0x00E0; //"dec fr"
				write_array(optemp1,line);
			}else bad_expr(line,linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"SUB",3)) //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SUB and remove spaces/tabs
			//assume first op is fr
			optemp1 = literalize(linemap[line],pos,line);
			for(; linemap[line][pos] != ',' ;pos++);
			if(linemap[line][pos] == ',')
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09 ;pos++) ; 
				if(linemap[line][pos] == 0x0D) bad_expr(line,linemap[line]);
				if(linemap[line][pos] == '#')  //confirms "sub fr,#lit"
				{                              //breaks into "mov w,#lit" "sub fr,w"
					optemp = mov_w_lit(linemap[line],pos,line,0);  //let function get lit
					write_array(optemp,line);
					optemp = sub_fr_w(linemap[line],pos,line,optemp1);
				}
				else if(linemap[line][pos] == 'W') //confirms "sub fr,w"
				{
					optemp = sub_fr_w(linemap[line],pos,line,optemp1);
				}
				else //assume it is "sub fr1,fr2"
				{   //breaks into "mov w,fr2" "sub w,fr1"
					optemp = mov_w_fr(linemap[line],pos,line,0);  //let function get optemp
					write_array(optemp,line);
					optemp = sub_fr_w(linemap[line],pos,line,optemp1);
				}
				write_array(optemp, line);
			}
			else bad_expr(line,linemap[line]);
		}
		else if(!strncmp((char *)linemap[line],"SWAP",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after SWAP and remove spaces/tabs
			optemp = literalize(linemap[line],pos,line);
			optemp <<= 11;
			optemp >>= 11;
			optemp += 0x03A0;
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"SZ",2)) //working
		{
			if((linemap[line][2] != 0x20) && (linemap[line][2] != 0x09) && (linemap[line][2] != 0x0D)) bad_expr(line,linemap[line]);  
			optemp = 0x0743;
			write_array(optemp, line);
		}
		else if(!strncmp((char *)linemap[line],"TEST",4)) //working
		{
			if((linemap[line][4] != 0x20) && (linemap[line][4] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 4; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after TEST and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //means "test w"
			{
				optemp = 0x0D00;
			}
			else //assume "test fr"
			{
				optemp = literalize(linemap[line],pos,line);
				optemp <<= 11;
				optemp >>= 11;
				optemp += 0x0220;
			}
			write_array(optemp,line);
		}
		else if(!strncmp((char *)linemap[line],"XOR",3))  //working
		{
			if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);  
			for(pos = 3; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //start off reading after XOR and remove spaces/tabs
			if(linemap[line][pos] == 'W')  //means "xor w,fr" or "xor w,#lit"
			{
				pos++;
				for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == '#')  //confims "xor w,#lit"
					{
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 8; 
						optemp >>= 8;
						optemp += 0x0F00;
					}
					else //assume "xor w,fr"
					{
						optemp = literalize(linemap[line],pos,line);
						optemp <<= 11;
						optemp >>= 11;
						optemp += 0x0180;
					}
				}
				else bad_expr(line,linemap[line]);
			}
			else //assume it is "xor fr,something"
			{
				optemp1 = literalize(linemap[line],pos,line);
				for(; linemap[line][pos] != ',' && linemap[line][pos] != 0x0D;pos++) ; //again ignore spaces
				if(linemap[line][pos] == ',')
				{
					pos++;
					for(; linemap[line][pos] == 0x20 || linemap[line][pos] == 0x09;pos++) ; //again ignore spaces
					if(linemap[line][pos] == 'W')  //means we have "xor fr,W"
					{
						optemp = xor_fr_w(linemap[line],pos,line,optemp1);  //we already got fr
					}
					else if(linemap[line][pos] == '#') //confirms "xor fr,#lit"
					{
						optemp = mov_w_lit(linemap[line],pos,line,0);  //let function get #lit
						write_array(optemp,line);
						optemp = xor_fr_w(linemap[line],pos,line,optemp1);  //er already got fr
					}
					else  //assume it is "xor fr1,fr2"  and we already have fr1
					{    //breaks into "mov w,fr2" "xor fr1,w"
						optemp = mov_w_fr(linemap[line],pos,line,0); //let function get fr2
						write_array(optemp,line);
						optemp = xor_fr_w(linemap[line],pos,line,optemp1);  //we have fr1
					}
						
				}else bad_expr(line,linemap[line]);
			}
			write_array(optemp,line);
		}
		else
		{
			fprintf(stderr,"Error: Line %d: Unknown instruction: %X\n",line + 1,(char *)linemap[line]);
			//exit(1);
		}
	}
	
	
	
	
	for(x = 0; x <= 15; x++)
	{
		id[x] = 0x0FFF;
	}
	writelines(opcodes, id, fuse, sx182028);
	
	
return 0;
}

int bad_expr(int line, char linemap[256])
{
	fprintf(stderr,"Error: Line %d: Bad expression: %s\n",line + 1,linemap);
	fcloseall();
	exit(1);
}

unsigned short int literalize(char linemap[256], int pos, int line)  //read in from the base identifier, till the , or EOL
{
	unsigned short int optemp = 0x00;
	int temp, multnum;
	
	if((linemap[pos] == '$'))// || ((linemap[pos] >= 0x30 && linemap[pos] <= 0x39) || (linemap[pos] >= 0x41 && linemap[pos] <= 0x46)))  //write in code to assume hex base of numbers
	{
		pos++;
		for(temp = 0; (linemap[pos] >= 0x30 && linemap[pos] <= 0x39) || (linemap[pos] >= 0x41 && linemap[pos] <= 0x46);temp++) pos++; //find out how long it is till the end
		for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) temp++; //again ignore spaces
		//printf("%c\n", linemap[pos]);
		if((linemap[pos] != 0x0D) && (linemap[pos] != ',') && (linemap[pos] != 0x0))  //make sure it is the first or last expression
		{
			bad_expr(line,linemap);
		}		
		pos--;
		
		multnum = 0;
		
		for(;temp;temp--)
		{
			
			if((linemap[pos] >= 0x30 && linemap[pos] <= 0x39))
			{
				optemp += ((linemap[pos] - 0x30) << (multnum * 4));
			}
			else if((linemap[pos] >= 0x41 && linemap[pos] <= 0x46))
			{
				optemp += ((linemap[pos] - 0x037) << (multnum * 4));
			}
			else if((linemap[pos] == 0x20) || (linemap[pos] == 0x09)) multnum--; //again ignore spaces
			else 
			{
				fprintf(stderr,"Error\n");
				exit(1);
			}
			pos--;
			multnum++;
			
		}
	}
	else if(linemap[pos] == '%')
	{
		pos++;
		for(temp = 0; ((linemap[pos] == 0x30) || (linemap[pos] == 0x31));temp++) pos++; 
		for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //again ignore spaces
		if((linemap[pos] != 0x0D) && linemap[pos] != ',')  //make sure it is the first or last expression
		{
			printf("fail\n");
			bad_expr(line,linemap);
		}
		pos--;
		
		multnum = 0;

		for(;temp;temp--)
		{
			for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;) 
			{
				//temp++; //again ignore spaces
				pos--;
			}
			if((linemap[pos] == 0x31))
			{
				optemp = (optemp + (1 << multnum));
			}
			else if(linemap[pos] == 0x30);
			else if((linemap[pos] == 0x20) || (linemap[pos] == 0x09) || (linemap[pos] == 0x0D)) ;
			else 
			{
				fprintf(stderr,"Error\n");
				exit(1);
			}
			pos--;
			multnum++;
		}
	}
	else if((linemap[pos] == '#') || ((linemap[pos] >= 0x30) && (linemap[pos] <= 0x39)))
	{
		if(linemap[pos] == '#')
		{
			pos++;
		}
		
		if((linemap[pos] == '$') || (linemap[pos] == '%'))
		{
			optemp = literalize(linemap,pos,line);
		}
		else
		{
			for(temp = 0; ((linemap[pos] >= 0x30) && (linemap[pos] <= 0x39));temp++) pos++; 
			for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) temp++; //again ignore spaces
			if((linemap[pos] != 0x0D) && linemap[pos] != ',')  //make sure it is the first or last expression
			{
				printf("fail\n");
				bad_expr(line,linemap);
			}
			pos--;
			
			multnum = 0;
				

			for(;temp;temp--)
			{
				
				if(((linemap[pos] >= 0x30) && (linemap[pos] <= 0x39)))
				{
					optemp += ((linemap[pos] - 0x30) * pow(10,multnum));  //this will add the int to the correct place
				}
				else if((linemap[pos] == 0x20) || (linemap[pos] == 0x09)) multnum--;
				else 
				{
					fprintf(stderr,"Error\n");
					exit(1);	
				}
				pos--;
				multnum++;
			}
			//if(optemp > 255) fprintf(stderr, "Warning: Line %d: Value larger than one byte, truncated\n",line + 1);
		}	
	}
	else bad_expr(line,linemap);
	return optemp;
}


int write_array(unsigned short int optemp, int line)
{
	if(opcodes[opline][oppos] != 0x0FFF)
	{
		fprintf(stderr,"Error: Line %d: Overwriting memory location\n",line + 1);
		exit(1);
	}
	opcodes[opline][oppos] = optemp;
	oppos++;
	if(oppos == 16)
	{
		opline++;
		oppos = 0;
	}
	return 0;
}

unsigned short int add_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)  //working
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	//pos++;
	
	//printf("%04X\n",optemp);
	optemp += 0x01E0;  //opcode for add fr,w
	//printf("%04X\n",optemp);
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int and_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)  //check
{
	if(!optemp)	optemp = literalize(linemap,pos,line);

	optemp <<= 11;
	optemp >>= 11;
	//for(;(linemap[pos] != ',');pos++) ;
	//pos++;
	optemp += 0x0160;  //opcode for and fr,w
	
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int mov_w_lit(char linemap[256], int pos, int line,unsigned short int optemp)  //working
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	
	optemp <<= 8;
	optemp >>= 8;
	//for(;(linemap[pos] != 0x0D);pos++) ;
	//pos++;
	
	optemp += 0x0C00;  //opcode for mov w,lit
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int mov_w_fr(char linemap[256], int pos, int line, unsigned short int optemp)  //working
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	//for(;(linemap[pos] != 0x0D);pos++) ;
	//pos++;
	
	optemp += 0x0200;  //opcode for mov w,fr
	//printf("%04X\n",optemp);
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int mov_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	//for(;(linemap[pos] != 0x0D);pos++) ;
	//pos++;
	
	optemp += 0x0020;  //opcode for mov w,fr
	//printf("%04X\n",optemp);
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int jmp(char linemap[256], int pos, int line, unsigned short int optemp)
{
	if(!optemp)optemp = literalize(linemap,pos,line);
	optemp <<= 7;
	optemp >>= 7;
	optemp += 0x0A00;
	return optemp;
}

unsigned short int sub_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	//for(;(linemap[pos] != 0x0D);pos++) ;
	//pos++;
	
	optemp += 0x00A0;  //opcode for sub w,fr
	//printf("%04X\n",optemp);
	//write_array(optemp, line);
	
	return optemp;
}

unsigned short int or_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)
{
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	
	optemp += 0x0120;  //opcode for or w,fr
	
	return optemp;
}

unsigned short int xor_fr_w(char linemap[256], int pos, int line, unsigned short int optemp)
{  
	if(!optemp) optemp = literalize(linemap,pos,line);
	optemp <<= 11;
	optemp >>= 11;
	//for(;(linemap[pos] != 0x0D);pos++) ;
	//pos++;
	
	optemp += 0x01A0;  //opcode for cor w,fr
	
	return optemp;
}

unsigned short int cjs(char linemap[256], int pos, int line, int cjnum)  //0 is cja, 1 is cjae, 2 is cjb, 
{																		//3 is cjbe, 4 is cje, 5 is cjne
	unsigned short optemp,optemp1;									   //6 is csa, 7 is csae, 8 is csb
	if( cjnum >= 0 && cjnum <= 11) //these have alot in common         //9 is csbe, 10 is cse, 11 is csne
	{
		
		for(pos = 4; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //start off reading after CJAE and remove spaces/tabs
		optemp1 = literalize(linemap,pos,line);  //get fr1
		for(;(linemap[pos] != ',')  && (linemap[pos] != 0x0D);pos++);
		
		if(linemap[pos] == ',')
		{
			pos++;
			for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //again ignore spaces
			if(linemap[pos] == '#')  //breaks down to "mov w,#lit^$FF" "add w,fr" "snc" "jmp addr9"
			{	
				
				if(cjnum == 0 || cjnum == 3 || cjnum == 6 || cjnum == 9)
				{
					optemp = literalize(linemap,pos,line);  //we have to do this manually since we dont 
					optemp ^= 0xFF;
					optemp <<= 8;
					optemp >>= 8;
					optemp += 0x0C00;  //"mov w,#lit^$FF
					write_array(optemp,line);
					optemp1 <<= 11;
					optemp1 >>= 11;
					optemp = (optemp1 + 0x01C0); //"add w,fr"
				}
				else if(cjnum == 1 || cjnum == 2 || cjnum == 4 || cjnum == 5 || cjnum == 7 || cjnum == 8 
				 || cjnum == 10 || cjnum == 11)  //"mov w,#lit"
				{
					optemp = mov_w_lit(linemap,pos,line,0);  //"mov w,#lit"
					write_array(optemp,line);
					optemp1 <<= 11;  //optemp holds fr
					optemp1 >>= 11;
					optemp = (optemp1 + 0x0080); //"mov w,fr-w"
				}
				else
				{
					fprintf(stderr,"Error\n");
					exit(1);
				}
				write_array(optemp, line); 
				
				if(cjnum == 0 || cjnum == 1 || cjnum == 8 || cjnum == 9)  //means it is cja or cjae third instrucion is "snc"
				{
					optemp = 0x0603;
					 //"snc"
				}
				else if(cjnum == 3 || cjnum == 2 || cjnum == 6 || cjnum == 7)  //means it is cjbe, third instruciont is "sc"
				{
					optemp = 0x0703;
					//"sc"
				}
				else if(cjnum == 4 || cjnum == 11)  //cje has third instruction of "snz"
				{
					optemp = 0x0643;
				}
				else if(cjnum == 5 || cjnum == 10) //cjne has third instrucion of "sz"
				{
					optemp = 0x0743;
				}
				else 
				{
					fprintf(stderr,"Something went wrong!\n");
					exit(1);
				}
				write_array(optemp, line);
				if(cjnum >= 0 && cjnum <= 5)
				{
					for(;(linemap[pos] != ',')  && (linemap[pos] != 0x0D);pos++);
					if(linemap[pos] == ',')
					{
						pos++;
						for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //again ignore spaces
						optemp = jmp(linemap,pos,line,0);
						/*optemp = literalize(linemap,pos,line); // finally get addr9
						optemp <<= 7;
						optemp >>= 7;
						optemp += 0x0A00;*/
					}
					else bad_expr(line,linemap);	
				}
			}
		
			else //assume it is "cja fr,fr2,addr9"  we have fr1 already and were at fr2
			{   //breaks into "mov w,fr1" "mov w,fr2-w" "sc" "jmp addr9"
				if(cjnum == 0 || cjnum == 3 || cjnum == 6 || cjnum == 9 || cjnum == 10)
				{
					optemp = mov_w_fr(linemap,pos,line,optemp1);  //we have fr1 already
					write_array(optemp, line); //"mov w,fr1"
					optemp = literalize(linemap,pos,line);  //get fr2
					optemp <<= 11;
					optemp >>= 11;
					optemp += 0x0080; //"mov w,fr2-w"
				}	
				else if(cjnum == 1 || cjnum == 2 || cjnum == 5 || cjnum == 4 || cjnum == 7 || cjnum == 8 || cjnum == 11)
				{
					optemp = mov_w_fr(linemap,pos,line,0);  //let function get fr2
					//"mov w,fr2"
					write_array(optemp, line);
					
					optemp1 <<= 11;
					optemp1 >>= 11;
					optemp = (optemp1 + 0x0080);  //"mov w,fr1-w"
				}					
				else
				{
					fprintf(stderr,"Error 3\n");
				}
				write_array(optemp, line);  
				
				if(cjnum == 0 || cjnum == 2 || cjnum == 7 || cjnum == 9)  //means we have cja, third instruction is sc
				{
					optemp = 0x0703;
					write_array(optemp, line); //"sc"
				}
				else if(cjnum == 3 || cjnum == 1 || cjnum == 6 || cjnum == 8) //means we have cjbe, third instrution is snc
				{
					optemp = 0x0603;
					write_array(optemp, line); //"snc"
				}
				else if(cjnum == 5 || cjnum == 10)
				{
					optemp = 0x0743; //"sz"
					write_array(optemp,line);
				}
				else if(cjnum == 4 || cjnum == 11)
				{
					optemp = 0x0643; //"snz"
					write_array(optemp,line);
				}
				else 
				{
					fprintf(stderr,"Something went wrong!\n");
					exit(1);
				}
				
				if(cjnum >= 0 && cjnum <= 5)
				{
					for(;(linemap[pos] != ',')  && (linemap[pos] != 0x0D);pos++);
					if(linemap[pos] == ',')
					{
						pos++;
						for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //again ignore spaces
						optemp = jmp(linemap,pos,line,0);
						/*optemp = literalize(linemap,pos,line); // finally get addr9
						optemp <<= 7;
						optemp >>= 7;
						optemp += 0x0A00;*/
					}
					else bad_expr(line,linemap);	
				}
			}
		}
		else bad_expr(line,linemap);
		if(cjnum >= 0 && cjnum <= 5) write_array(optemp, line); //skip commands dont need the jump written
	}
	return 0;
}

unsigned short int j(char linemap[256], int pos, int line, int jnum)  //jnum 0 is jb, 1 is jc, 2 is jnb, 3 is jnc, 4 is jnz, 5 is jz
{
	unsigned short int optemp;
	//if((linemap[line][3] != 0x20) && (linemap[line][3] != 0x09)) bad_expr(line,linemap[line]);
	for(pos = 3; linemap[pos] == 0x20 || linemap[pos] == 0x09 /*|| linemap[line][pos] == 0x0D*/;pos++) ; //start off reading after and remove spaces/tabs
	//if(linemap[line][pos] == 0x0D) bad_expr(line,linemap[line]);  see if this is really necessary
	//assume we have addr 9
	if(jnum == 1) optemp = 0x0603; //"snc"
	else if(jnum == 3) optemp = 0x0703;  //"sc"
	else if(jnum == 4) optemp = 0x0743;  //"sz"
	else if(jnum == 5) optemp = 0x0643;  //"snz"
	else
	{
		fprintf(stderr,"Error\n");
		exit(1);
	}
	write_array(optemp, line);
	optemp = jmp(linemap,pos,line,0);
	write_array(optemp,line);
	return 0;
}

//working
unsigned short int dobit(char linemap[256], int pos, int line) //takes in starting at op.bit, 
{															  //will find end and return optemp in forrect format
	unsigned short int optemp, optemp1 = 0;
	char templine[256];
	int temp, x;

	memset(templine,0x00,245);
	if(linemap[pos] == '/') pos++;	
	for(temp = pos; linemap[temp] != '.' ;temp++) ;
	
	strncpy(templine,linemap + pos,(temp - pos));	
	optemp1 = literalize(templine, 0,line);  //this will give us the fr
	pos = temp + 1;  //get back the the "." and move one space more
	for(; linemap[pos] == 0x20 || linemap[pos] == 0x09;pos++) ; //again ignore spaces
	optemp = linemap[pos];
	optemp <<= 13;  //this clears optemp and put the .bit at byte boundry
	optemp >>= 8;
	optemp += optemp1;

	return optemp;
}
